      Program histo_knuth
! calcule un histogramme
! calcule le nombre optimal de d'intervalles pour construire l'histogramme
! gestion des incertitudes Ref: Knuth (2006) arXiv:physics/0605197
! germain_vallverdu@yahoo.fr
      Implicit None
      integer::i,j,npas,ntot,optbin,metho
      double precision::xx,bid,xmin,xmax,xm,x2
      double precision::pas,val
      double precision::vol,hnk,som
      character(len=50)::fichier,titre
      double precision,dimension(:),allocatable::x
      double precision,dimension(:,:),allocatable::h

      write(*,*)
      write(*,*)"     ***  histogramme  *** "
      write(*,*)
      write(*,*)"Méthode : 1 = Choix du nombre d'intervalles"
      write(*,*)"          2 = Optimisation du nombre d'intervalles"
      read(*,*)metho
      write(*,*)
      write(*,*)"Format du fichier de donnees"
      write(*,*)"     ligne  1 = un titre"
      write(*,*)"     colone 1 = un indice (non utilise)"
      write(*,*)"     colone 2 = variable"
      write(*,*)
      write(*,*)"donner le nom du fichier"
      read(*,*)fichier
      write(*,*)"_____________________________________________________"
      write(*,*)

! ========== lecture des donnees
      write(*,*)"lecture du fichier = "//trim(fichier)
      open(10,file=trim(fichier))
      read(10,"(a)",err=1000)titre
      i=0
      xmax=-1.d-70 
      xmin= 1.d+70 
      xm = 0.d0
      Do
        i=i+1
        read(10,*,end=100,err=1000)bid,xx
        xmin = min(xx,xmin)
        xmax = max(xx,xmax)
        xm = xm + xx
      Enddo
100   close(10)
      ntot=i-1
      xm = xm/dble(ntot)

      write(*,"(a,i9)")   "nombre de points lus     : ",ntot
      write(*,*) 
      write(*,"(a,F12.6)")"valeur min               : ",xmin
      write(*,"(a,F12.6)")"valeur max               : ",xmax

      allocate( x(1:ntot) )

      x(:)=0.d0
      open(10,file=trim(fichier))
      read(10,*)
      Do i=1,ntot
        read(10,*)bid,x(i)
        x2 = x2 + (x(i) - xm)**2
      Enddo
      close(10)
      x2 = sqrt(x2/dble(ntot))
      write(*,"(a,F12.6)")"moyenne                  : ",xm
      write(*,"(a,F12.6)")"ecart type               : ",x2

! ========== construction de l'histogramme
      Select Case(metho)
        case(1) ! classique
          write(*,*)
          write(*,*)"Donner le nombre d'intervalles voulus"
          read(*,*)npas
        case(2) ! optimisation du nombre d'intervalles
          npas = optbin(ntot,x(:),xmin,xmax)
        case default
          STOP "Choix de la méthode incorrect"
      End Select

      pas = (xmax - xmin) / dble(npas)

      write(*,*)
      write(*,"(a,F12.6)")"pas de l'histogramme     : ",pas
      write(*,"(a,i9)")   "nombre de d'intervalles  : ",npas
      write(*,*)

      allocate(h(1:npas,1:3))

      call histogramme(ntot,x(:),xmin,xmax,npas,h(:,:),npas)
 
! ========== calcul des incertitudes => article Knuth
      val = dble(npas) / (xmax - xmin)
      Do j=1,npas
        hnk = h(j,2)

        h(j,2) = val*( hnk + 0.5d0 )/( dble(ntot) + dble(npas)/2.d0 )

        h(j,3) = val * sqrt(                                           &
            ( hnk + 0.5d0 )*( dble(ntot) - hnk + dble(npas-1)/2.d0 ) / &
            ( dble(ntot+1) + dble(npas)/2.d0 ) /                       &
            ( dble(ntot  ) + dble(npas)/2.d0 )**2 )
      Enddo

! ========== calcul de la norme => methode des rectangles
      som = 0.d0
      Do j=1,npas
        som = som + h(j,2)*pas
      Enddo
      write(*,"(a,F12.6)")"Norme de l'histogramme   : ",som

! ========== impression
      open(10,file=trim(fichier)//".histo")
      write(10,"(a)")"# "//titre
      write(10,"(a)")"#      x         histo"// &
                     "       Delta h   Delta h / h"
      Do j=1,npas
        write(10,"(4F12.6)")h(j,1),h(j,2),h(j,3),h(j,3)/h(j,2)
      Enddo
      close(10)

      write(*,*)
      write(*,*)"histogramme ecrit dans : "//trim(fichier)//".histo"
      write(*,*)
      write(*,*)"fin"
      STOP

1000  STOP "erreur de lecture du fichier de données"
     
      End

! ==============================================================================

      Subroutine histogramme(ntot,x,xmin,xmax,ndim,h,npas)
      Implicit None
      integer,intent(in)::ntot,npas,ndim
      double precision,dimension(1:ntot),intent(in)::x
      double precision,dimension(1:ndim,1:3),intent(out)::h
      double precision,intent(in)::xmin,xmax
      integer::i,j
      double precision::vali,valf,val0,val,pas

      pas = (xmax - xmin) / dble(npas)
      val0 = xmin + pas/2.d0

      h(:,:)=0.d0
      Do j=1,npas   !histo
        val = val0 + pas*dble(j-1)
        vali = val - pas/2.d0
        valf = val + pas/2.d0
        h(j,1)= val    
        Do i=1,ntot,1
          If ( x(i) >= vali .and. x(i) < valf ) then
            h(j,2) = h(j,2) + 1.d0
          Endif
        Enddo  !ntot
      Enddo  !histo

      end

! ==============================================================================

      Function optbin(ntot,x,xmin,xmax)
      Implicit None
      integer::optbin
! recherche du nombre idéal de bin de l'histogramme
! optimisation d'une fonction
! Ref : K.H. Kuhn (2006) arXiv:physics/0605197 
      integer,intent(in)::ntot
      double precision,dimension(1:ntot),intent(in)::x
      double precision,intent(in)::xmin,xmax
      integer,parameter::ndim=100
      integer::m_min,m_max,i,m_opt,npas,j
      double precision::gammln,val,vmax_logp,pas
      double precision,dimension(1:ndim,1:3)::h

      m_min = 2
      m_max = min(ntot,ndim)

      vmax_logp=-1.d300
      Do npas= m_min, m_max

         call histogramme(ntot,x(:),xmin,xmax,ndim,h(:,:),npas)
 
         val= dble(ntot)*log( dble(npas) )            &
              + gammln( dble(npas)/2.d0 )             &
              - gammln( dble(ntot) + dble(npas)/2.d0) &
              - dble(npas)*gammln(0.5d0)

         Do i= 1,npas
            val= val + gammln( h(i,2) + 0.5d0 )
         Enddo

         If( val > vmax_logp ) then
            m_opt = npas
            vmax_logp = val
         Endif
 
      Enddo

      optbin=m_opt

      end

! ==============================================================================

      Function gammln(x)
! calcul de la fonction ln( Gamma(x) ) 
! dans le cas où x = n (n entier)
! dans le cas où x = n + 1/2 (n entier)
! formules issues de Abramowitz and Stegun (1972)
      Implicit None
      double precision,intent(in)::x
      double precision::pi
      integer::i,k,m
      double precision::gammln
      pi = acos(-1.d0)
      k = nint(2.d0*x)
      If    ( mod(k,2) == 0 ) Then
        m = nint(x)
        gammln = 0.d0
        Do i=1,m-1
          gammln = gammln + log( dble(i) )
        Enddo
      ElseIf( mod(k,2) == 1 ) Then
        m = ( nint(2.d0*x) - 1 ) / 2
        gammln = 0.5d0*log(pi) - dble(m)*log(2.d0)
        Do i=1,m
          gammln = gammln + log( dble( 2*i-1 ) )
        Enddo
      Else
        STOP "## gammln n ou n+1/2 non identifie"
      Endif
      End

